#!/usr/bin/env python
doc="""
*********************** pti2fluorescence2D ****************************
Convert triangular 2D fluorescence data exported from PTI FelixGX to 
the plain-text format for Fluorescence2D 

6/23/2013
v.2 by Evgenii Kovrigin, Marquette University


This utility makes PTI 2D triangles usable in MATLAB (converts them to 2D rectangle plane).
I suppose that the file contains the a series of emission scans with incrementing 
starting position and excitation wave length as:
excitation at: 200nm    emission range: 210-600
excitation at: 210nm    emission range: 220-600
excitation at: 220nm    emission range: 230-600
etc.

The result is a triangular data table where the raw detector data are 
followed by corrected data:

Detector            Detector (COR)
200 data 210 data  220 data  200 data 210 data 220 data 
... data ... data  ... data  ... data ... data ... data
... data ... data  600 data  ... data ... data 600 data
... data 600 data            ... data 600 data
600 data                     600 data 

IMPORTANT: The first trace must be the longest one! That is we are INCREMENTING
excitation wavelength and the blue edge of emission range.


To export this data one should use the toolbar function of FelixGX "Export data" and
have only "Detector" data sets selected while exporting.

This utility  fills in some constant value  for frequencies in the front the shorter traces to create
rectangular tables and exports them as a text file.


IMPORTANT: If the list datasets of FelixGX contains traces with the additional suffix (2),
you must delete them before export. Othewise, "IndexError: list index out of range" will
be issued by Python. Typically, such trace result from previous measurements left over 
in the experiment before starting collection of the 2D data. You should "Close record" 
before starting a macro to avoid this problem.


Usage:   pti2fluorescence2D.py  -raw|-corr 2D_data_in_triangle_form.txt 
                where -raw switch indicates a need to export raw data
                while -corr switch will direct export of corrected data
                
Bugs and fixes:
3/8/2014 - dependence on mylib.py removed
3/9/2014 - utility renamed to pti2fluorescence2D.py
3/19/2014 - extraction of data from PTI file re-written to make fewer assumptions
            and more internal consistency checks.
4/28/2014 - the code re-written as a single class

*****************************************************************
"""


import os, sys, string,  copy 
ZERO_VALUE='zero'  # placeholder of a value to be placed in empty triangle

#----------- CLASS DESCRIPTIONS ---------------
class PTI2Fluorescence2D:
    'converter for triangular or trapezoid 2D data'
    
    pti_file=[];  # storage for the raw PTI data file
    
    # Structure of the PTI Group file (see PTI manual for details)
    pti_header_length=6  # number of strings in the header
    pti_footer_length=1  # number of strings in the footer
    excitation_frequency_header_line=4 # this is where frequencies are
    
    
    # Data storage
    pti_header=[]
    pti_data=[]
    pti_footer=[]
    Emission_wavelengths=[] # a vector of all emission frequencies
    Excitation_wavelengths=[] # a vector of all excitation wavelengths
    intensity_data=[]  # matrix of intensity values 
    
    # Flag -- what part was extraced COR or raw
    data_part='error_not_converted_yet'
    
    
    
    
    #========== CONSTRUCTOR =================================
    def __init__(self, fname):
        " creates class object and loads the 2D data file"

        # check if file exists
        if not os.path.isfile(fname): 
          self.error("\n\n\nError!\nCannot open '%s' to read!!!\n\n\a\a\a" % fname)

        # read PTI file        
        input=open(fname,'r')	
        L=input.readlines()
        M=[]
        for line in L:
                if line.strip()!='':
                    M.append(line)
        input.close()
        self.pti_file=M
        


        # Slice the input file in portions
        self.pti_header=self.pti_file[:self.pti_header_length]
        self.pti_data=self.pti_file[self.pti_header_length:-self.pti_footer_length] 
        self.pti_footer=self.pti_file[-self.pti_footer_length:]

        
        # Check if the data are of a correct kind
        if '<Group>' not in self.pti_header[0]:
            self.error("Error! The data should be an exported from FelixGX via 'Export Group' command!\nI do not find <Group> keyword in the first line of '%s'!\n" % self.pti_fname)


        # ===== Extract excitation wavelengths from the header. Check presence of corrected data ======
        # Rely on the condition that excitation wavelength is always followed by a colon:
        # EXCITATION_FREQ:MIN-MAX
        words=self.pti_header[self.excitation_frequency_header_line].split(':')[:-1] 
        # INTERNAL CHECK: the keyword [COR] must be present in the last half of the array
        last_word=words[-1].split()
        keyword_cor='[COR]'
        if keyword_cor not in last_word:
            self.error('\nDATA FORMAT ERROR: The data must contain raw and corrected traces with the label [COR]!\nSee the right half of the line %d of %s.'%(
                 self.excitation_frequency_header_line+1, fname))
        # take left half
        left_half= words[:len(words)/2]
        # extract excitation wavelengths (rightmost word)
        for item in left_half:
          lambda_ex=string.split(item)[-1]
          self.Excitation_wavelengths.append(lambda_ex)

        print "Excitation wavelenghts detected:"
        print self.Excitation_wavelengths    
        
        
        
        
        
        
    
    # ===  CONVERT TRIANGULAR DATA TO RECTANGULAR FORMAT ===
    def convert(self, flag):
        "flag: -corr or -raw to choose which part of the data to use"
                

        # ===== REFORMATTING THE DATA =======
        # Prepare matrices for MATLAB in the following format:
        # Emission_wavelengths --- a vector of all emission frequencies
        # Excitation_wavelengths --- a vector of all excitation wavelengths
        # Intensity_matrix --- a rectangular matrix of intensities, where 
        #     the missing values in the second half of the triangle 
        #     a replaced with the value, which is the zeros. 
        #
        # The goal is to reshuffle the existing triangle:
        #    Ex1     Ex2    Ex3
        #  Em1 Y1  Em2 B2  Em3 D3
        #  Em2 Y2  Em3 B3
        #  Em3 Y3  
        #
        #  to form the following rectangle
        #
        #       Em1 Em2 Em3 ...
        # Ex1    Y1  Y2  Y3 ...
        # Ex2    0   B2  B3 ...
        # Ex3    0   0   D3
        # ...



        # Read Data size 
        N_traces_line=3
        N_experiments=int(self.pti_header[N_traces_line-1])
        #print N_experiments


        #=== RECREATE A RECTANGLE ===
        # replace all empty items (where only tabs were) with zeros
        new_pti_data=[] # array of arrays
        for line in self.pti_data:
            split_data=line.split('\t')
            new_split_data=[] # array for each line
            for item in split_data:
                item=string.strip(item)  # get rid of '/r/n'
                if item=='':
                    new_split_data.append(ZERO_VALUE)
                else:
                    new_split_data.append(item)
            new_pti_data.append(new_split_data)
            
        # DEBUG: To see the zero-replaced table turn this output on:
        show_new_pti_data='n' 
        if show_new_pti_data=='y' :
          for line in new_pti_data:
            for item in line:
                print '%10s\t' % item,
            print '\n'
          print '\n'
          print '\n'
          print '\n'
            

        # Transpose the dataset
        transposed_pti_data=[]
        for column_index in range(len(new_pti_data[0])):
            new_column=[]
            for row_index in range(len(new_pti_data)):
                item=new_pti_data[row_index][column_index]
                new_column.append(string.strip(item))
            transposed_pti_data.append(new_column);
            
        # DEBUG: To see the transposed table turn this output on:
        show_transposed_pti_data='n' 
        if show_transposed_pti_data=='y' :
          for line in transposed_pti_data:
            for item in line:
                print '%10s\t' % item,
            print '\n'
          print '\n'
          print '\n'
          print '\n'

        # Shift values forward to the end into empty cells; fill with zeros
        # new empty: 
        # [ 1 2 3 0 0 ] => [ 0 0 0 1 2 3 ]        

        shifted_pti_data=[]
        for line in transposed_pti_data:
            #print line  
            new_line=[]
            # begin from end. Count zeros:
            counter1=0
            while line[-1-counter1]==ZERO_VALUE:
                #print counter1, line[-counter1]
                new_line.append(ZERO_VALUE)
                counter1+=1
            # now add all non-zero data from the front
            counter2=0
            while line[counter2]!=ZERO_VALUE :
                #print counter2, line[counter2-1]
                new_line.append(line[counter2])
                counter2+=1
                if len(line)==counter2:
                    break

            #print new_line
            shifted_pti_data.append(new_line)
            
        # DEBUG: To see the shifted table turn this output on:
        show_shifted_pti_data='n' 
        if show_shifted_pti_data=='y' :
          for line in shifted_pti_data:
            for item in line:
                print '%10s\t' % item,
            print '\n'
          print '\n'
          print '\n'
          print '\n'


        # Next steps 
        # 1. split data exactly in half (top - raw portion,
        # bottom - corrected portion) 

        # 2. correctly extract excitation wavelengths
        # 3. take emission wavelengths from the top line of the shifted dataset

        # INTERNAL CONTROL: Number of lines must be double of number of traces 
        # given in the file header. 
        if len(shifted_pti_data)!=N_experiments*2:
            self.error( "\nINTERNAL ERROR: Number of traces in the data is %s. Expected %s" %(
            len(shifted_pti_data)/2,N_experiments))


        # Split data into two halves (left and right) based
        # on repetition of the wavelength values of the first dataset
        # in the X column for raw and corrected detector data
        raw_data=shifted_pti_data[:N_experiments]  # top half
        corrected_data=shifted_pti_data[N_experiments:] # bottom half

        # choose current dataset (half)
        if flag=='-raw':
            data=raw_data
            self.data_part='raw'
            
        if flag=='-corr':
            data=corrected_data
            self.data_part='corr'

        # DEBUG: To see the shifted table turn this output on:
        show_data='n' 
        if show_data=='y' :
          for line in data:
            for item in line:
                print '%10s\t' % item,
            print '\n'
          print '\n'
          print '\n'
          print '\n'




        # Extract emission frequencies from the first words of all lines of the data
        self.Emission_wavelengths=data[0]
        print "Emission wavelenghts detected:"
        print self.Emission_wavelengths    


        # Extract pure intensity data (skip intevening emission wavelength lines)
        flag=0
        for line in data:
            if flag:
                # replace zeros with zero values
                new_line=[]
                for item in line:
                    if item==ZERO_VALUE:
                        new_line.append('0')
                    else:
                        new_line.append(item)
                        
                self.intensity_data.append(new_line)
                flag=0
                continue
            else:
                flag=1
                
        # DEBUG: To see the  table turn this output on:
        show_intensity_data='n' 
        if show_intensity_data=='y' :
          for line in self.intensity_data:
            for item in line:
                print '%10s\t' % item,
            print '\n'
          print '\n'
          print '\n'
          print '\n'

    
    
            
                    
                                    

    #===== OUTPUT CONVERTED DATA FOR MATLAB ====
    def output(self, pti_fname):
        "export converted data for processing in MATLAB"    

        # output dataname
        dirname=pti_fname+'.'+self.data_part
        if  os.path.isdir(dirname):
            os.system('rm -r "%s"' % dirname)
        os.mkdir(dirname)

        fidEm=open(os.path.join(dirname,'emission_wavelengths_Xaxis.txt'),'w')
        fidEm.writelines(string.join(self.Emission_wavelengths,'\t'))
        fidEm.close()

        fidEx=open(os.path.join(dirname,'excitation_wavelengths_Yaxis.txt'),'w')
        fidEx.writelines(string.join(self.Excitation_wavelengths,'\n'))
        fidEx.close()

        fidInt=open(os.path.join(dirname,'spectral_intensity.txt'),'w')
        for line in self.intensity_data:
            for item in line:
                fidInt.write(item+'\t')
            fidInt.write('\n')
        fidInt.close()

        print "Conversion for MATLAB completed.\nData were written into the folder: '%s'\n" % dirname







    
    #================= ERROR HANDLER ================================
    def error(self, error_message):
        "generic error handler for the class"
        print error_message
        sys.exit()


#============ END OF PTI2Fluorescence2D CLASS DEFINITION ========================









#------------ SCRIPT BEGINS -------------------

# Check if arguments are given
if len(sys.argv) !=3:
    print doc
    sys.exit()

if sys.argv[1]!='-raw' and sys.argv[1]!='-corr':
    print "ERROR! in the -raw or -corr switch!"
    sys.exit()

# Extract command line arguments
flag=sys.argv[1]
pti_fname=sys.argv[2]

# Load data
pti2Ddata =  PTI2Fluorescence2D(pti_fname);

# Convert data
pti2Ddata.convert(flag)

# Export data
pti2Ddata.output(pti_fname)


























